try:
    from fastapi import APIRouter, Header, Request
    from starlette.exceptions import HTTPException as StarletteHTTPException
    from pydantic import BaseModel
    from typing import Optional, Union

    router = APIRouter(include_in_schema = False)

    def abort(code: int) -> StarletteHTTPException:
        raise StarletteHTTPException(status_code=code)

    class VoteInternal(BaseModel):
        """
            Represents a internal IBL Vote. IBLPy will make this a Vote class
        """
        timeStamp: str
        userID: str
        botID: str
        type: str
        timestamp: Optional[int] = None

    class Vote():
        """
            Represents a vote on IBL

            :param bot_id: The Bot ID of the vote

            :param user_id: The ID of the user who voted for your bot. In test mode, this will be 0

            :param test: Whether this is a test webhook or not

            :param timestamp: The timestamp (epoch) when the vote happened
        """
        def __init__(self, bot_id: int, user_id: int, test: bool, timestamp: int):
            self.bot_id = bot_id
            self.user_id = user_id
            self.test = test
            self.timestamp = timestamp

    @router.post("/")
    async def iblpy_webhook(vote_internal: VoteInternal, Authorization: str = Header("INVALID_SECRET")):
        if secret is None or secret == Authorization:
            pass
        else:
            return abort(401)
        timestamp = int(vote_internal.timeStamp)
        
        if vote_internal.type.lower() == "test":
            bot_id = botcli.id
            user_id = 0
            test = True
        else:
            bot_id = int(vote_internal.botID)
            user_id = int(vote_internal.userID)
            test = False
        vote = Vote(bot_id = bot_id, user_id = user_id, test = test, timestamp = timestamp)
        return await wh_func(vote, secret)

except:
    pass




import IBLPy.config as cfg
from IBLPy import api_modes, aiohttp
from typing import Union, Optional

class InvalidMode(Exception):
    """Raised when you don't have the required mode (package) to perform the action such as trying to do an asynchronous API request without having aiohttp_requests installed or trying to do a webhook without fastapi+uvicorn"""
    def __init__(self, mode):
        if mode == "async":
            super().__init__("In order to use IBLPy asynchronous API requests, you must have aiohttp_requests installed")
        elif mode == "fastapi":
            super().__init__("In order to use IBLPy webhooks, you must have fastapi and uvicorn installed")

class Ratelimit(Exception):
    """Raised when you are being ratelimited by IBL. The ratelimit for posting stats is 3 requests per 5 minutes and is unknown/variable for getting stats from the API"""
    def __init__(self):
        super().__init__("You are being ratelimited by the Infinity Bots (IBL) API. For future reference, the ratelimit for posting stats is 3 requests per 5 minutes and is unknown/variable for getting stats from the API!")

class APIResponse():
    """
        IBLAPIResponse represents an API response in the IBLPy library
        
        :param res: This is the raw response from the API. 
            This will be a aiohttp ClientResponse

        :param success: Whether the API response has succeeded or not (status less than 400)
        
        :param error_msg: The error message reported by the Infinity Bot List API

        :param message: Any messages returned by the API in the message field. Can be None if there are no messages

        :param json: The JSON object sent by the API

        :param status: The status code of the HTTP response received from the API
    """
    def __init__(self, *, res: aiohttp.ClientResponse, json: dict):
        self.res = res
        self.success = True if res.status < 400 else False
        self.message = json.get("message")
        self.error = json.get("error")
        self.json = json
        self.status = res.status

class BaseUser():
    """
        This is a base user on IBL from which all bots and users extend from
    """
    def __init__(self, id, json):
        self.__dict__.update(**json)
        self.id = id
        self.clean()
    
    def dict(self) -> dict:
        """Returns the class as a dict using the dict dunder property of the class"""
        return self.__dict__

    def __str__(self) -> str:
        """Returns the name of the bot or user"""
        return self.name

    def __int__(self) -> int:
        """Returns the bot or user ID"""
        return self.id
    
    def clean(self):
        """Cleans up all the ugly stuff from the IBL API"""
        for k, v in self.__dict__.items():
            self.__dict__[k] = self._cleaner(attr=v)
    
    def _cleaner(self, attr: str):
        if isinstance(attr, str):
            if attr == "none":
                return None
            elif attr == "false":
                return False
            elif attr == "true":
                return True
            elif attr.isdigit():
                return int(attr)
        
        elif isinstance(attr, dict):
            _tmp = {}
            for k, v in attr.items():
                _tmp[k] = self._cleaner(attr=v)
            return _tmp
        
        return attr
    
class Bot(BaseUser):
    """
        IBLBot is internally a part of the base_fn module (which provides all of IBLPy's base classes and functions). It represents a bot on IBL. The exact parameters of an IBLBot may change and IBLPy is designed to handle such changes automatically. Here are the parameters that we know of right now:
    """
    def __init__(self, id, json):
        super().__init__(id, json)
        
class User(BaseUser):
    """
        IBLUser is internally a part of the base_fn module (which provides all of IBLPy's base classes and functions). It represents a user on IBL. The exact parameters of an IBLUser may change and IBLPy is designed to handle such changes automatically. Here are the parameters that we know of right now:
    """
    def __init__(self, id, json):
        super().__init__(id, json)     



import IBLPy.config as cfg
import IBLPy.http as _http
from IBLPy.classes import InvalidMode, IBLAPIResponse
from IBLPy import api_modes, fastapi, uvicorn, ws, discord
from loguru import logger
from typing import Optional, Awaitable
import os
import asyncio
import time

class BotClient():
    """
        Initialize a Infinity Bots List (IBL) Bot Client. You can use this to get/post bot stats and fetch a bot!
            
        :param id: The Bot ID you wish to use with the IBL API

        :param api_token: The API Token of the bot. You can find this by clicking API Token under the "Owner Section". 
            This is optional however you will not be able to post stats if you do not pass a API Token
    """

    def __init__(self, bot_id: int, api_token: Optional[str] = ""):
        self.bot_id = bot_id
        self.http = _http.BotHTTP(bot_id)
        
        if api_token:
            self.http.login(api_token)

            
    async def set_stats(self, guild_count: int, shard_count: Optional[int] = None) -> IBLAPIResponse:
        """
            Posts bot stats to the IBL API asynchronously
            
            :param guild_count: Amount of servers your bot is in

            :param shard_count: Amount of shards your bot is in. This is optional

            :return: This will always be returned unless something goes wrong, in which case you will get an exception
            :rtype: IBLAPIResponse 
        """
        return await self.http.set_stats(guild_count, shard_count)
  
    
    async def get_bot(self):
        """
            Asynchronously get a bot. This does not take any parameters.

            :return: If the bot was not found or if IBLPy could not parse the JSON request for any reason
            :rtype: None

            :return: This will be returned if you are ratelimited
            :rtype: IBLAPIResponse

            :return: This is the bot object returned from the API
            :rtype: IBLBot
        """
        return await self.http.get_bot()

class UserClient():
    """
        Initialize a Infinity Bots List (IBL) User Client. You can use this to get user stats!
            
        :param user_id: The User ID you wish to use with the IBL API
    """
    def __init__(self, user_id: int):
        self.user_id = user_id
        self.http = _http.UserHTTP(user_id)

    async def get_user(self):
        """
            Get a user. This does not take any parameters.

            :return: If the user was not found or if IBLPy could not parse the JSON request for any reason
            :rtype: None

            :return: This will be returned if you are ratelimited
            :rtype: IBLAPIResponse

            :return: This is the user object returned from the API
            :rtype: IBLUser
        """
        return await self.http.get_user()


async def _default_get_counts(ap):
    """Returns guild and shard count for default autoposter"""
    return {
        "guild_count": len(ap.discli.guilds),
        "shard_count": len(ap.discli.shards) if ap.sharding else None
    }
    
class AutoPoster():
    """
        This is IBL Auto Poster. It will post stats for you on an interval you decide
            and it will then run a function of your choice (if requested) after posting each time.

        You can stop a Auto Poster by calling the ``stop`` function.
        
        :param get_counts: A async function that takes in the AutoPoster object and then
            returns the dict {"guild_count: guild count, "shard_count": shard_count} for the autoposter.
            A default value for this is provided for ease of use

        :param interval: How long to wait each time you post. It is required to use at least 5 minutes which is 300 seconds.
            Specifying a value below 300 seconds will set the interval to 300 seconds anyways

        :param botcli: The IBL BotClient of the bot you want to post stats for. 
            It must have the API token set either during initialization or afterwards by setting the api_token parameter

        :param discli: The discord.Client (or like interfaces like discord.Bot/discord.AutoShardedBot) of the bot you want to post stats for.
            If you wish to add shard counts, please give a discord.AutoShardedClient and pass sharding = True or use get_counts

        :param on_post: The function to call after posting. Set to None if you don't want this.
            This function needs to accept three arguments, the guild count sent, the shard count set 
            and the response (which will either be a IBLAPIResponse, a IBLAPIRatelimit or None). 
            Shard count will be None if sharding is disabled and/or get_count provided doesn't support it

        :param sharding: Whether we should post sharding as well. Requires a discord.AutoShardedClient
    """
    def __init__(
        self, 
        interval: int,
        botcli: BotClient, 
        discli: discord.Client, 
        sharding = None, 
        on_post = None,
        get_counts: Awaitable = _default_get_counts,
    ):

        if interval > 300:
            self.interval = interval
        else:
            self.interval = 300
        self.get_counts = get_counts
        self.botcli = botcli
        self.discli = discli
        self.on_post = on_post
        self.sharding = sharding
        self.keep_running = True
        self.task = None

    async def autoposter(self):
        """This is the autoposting function that is run by the ``start`` function here (using asyncio.create_task)"""
        while True:
            if not self.keep_running:
                self.keep_running = True
                logger.info("IBLPy: AutoPoster has exited successfully!")
                return

            if self.interval < 300:
                self.interval = 300

            try:
                
                try:
                    counts = await self.get_counts(self)
                except Exception as exc:
                    logger.warning("Using self in get_counts did not work, trying no args...")
                    counts = await self.get_counts()
                    
                res = await self.botcli.set_stats(counts["guild_count"], counts["shard_count"])
                if self.on_post:
                    await self.on_post(counts["guild_count"], counts["shard_count"], res)
                logger.success(f"IBLPy: Stats posted successfully at {time.time()}")
            except Exception as ex:
                logger.error(f"IBLPy: Could not post stats because: {ex}")
            await asyncio.sleep(self.interval)

    def start(self):
        """Starts the auto poster using asyncio.create_task and returns the created task. This must be done in the on_ready function in discord.py"""
        self.task = asyncio.create_task(self.autoposter())
        return self.task

    def stop(self):
        """
            Stops the auto poster by setting the ``keep_running`` attribute to False. Since the exact name may change, it is recommended to use ``stop`` instead of manually setting the attribute.

            The autoposter will only stopped after the interval specified

            This will return the autoposter asyncio task
        """
        self.keep_running = False
        print("IBLPy: AutoPoster is stopping... Waiting for it to next wake up to post stats")
        return self.task

class Webhook():
    """
        This is an IBLPy webhook. It takes a BotClient, your webhook secret and the awaitable/asyncio corouting to call after getting a vote. To start your webhook:

            Use start_ws_task if you are running this in discord.py. make sure you start the webhook in your on_ready event (at the very end of it since it's blocking)

            Use start_ws_normal if you are running this seperately from discord.py. It will be run using asyncio.run instead of asyncio.create_task

            Use the start_ws coroutine if you want to run the webhook yourself using asyncio

            Use create_app if you want to get the FastAPI ASGI app and deploy the webhook using your own server. Uvicorn still needs to be installed even if you use a different ASGI server
    """
    def __init__(self, botcli: BotClient, coro: object, secret: str = None):
        if "webhook" not in api_modes:
            raise InvalidMode("fastapi")
        self.botcli = botcli
        self.secret = secret
        self.coro = coro

    def start_ws_task(self, route, port = 8012):
        """
            Start webhook using asyncio.create_task (discord.py). Use this in a discord.py on_ready event (at the very end of it since it's blocking).
        """
        asyncio.create_task(self.start_ws(route, port = port))

    def start_ws_normal(self, route, port = 8012):
        """
            Start webhook using asyncio.run for normal non-discord.py usecases
        """
        asyncio.run(self.start_ws(route, port = port))

    def create_app(self, route):
        """Creates a FastAPI ASGI app and returns it so you can deploy it how you want to deploy it"""
        ws.wh_func = self.coro
        ws.botcli = self.botcli
        ws.secret = self.secret
        app = fastapi.FastAPI(docs_url = None, redoc_url = None)
        app.include_router(
            router = ws.router,
            prefix=route,
        )
        return app

    async def start_ws(self, route, port = 8012):
        """
            Coroutine to start webhook using any asyncio method you want.
        """
        server = uvicorn.Server(uvicorn.Config(self.create_app(route), host = "0.0.0.0", port = port))
        await server.serve()
        try:
            os._exit()
        except:
            os._exit(0)
